#include <FS.h> 
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <DNSServer.h>
#include <WiFiManager.h>
#include <Ticker.h>
#include <ESP8266mDNS.h>
#include <ArduinoJson.h>
#include <Scheduler.h>
#include <Task.h>



#define STATE_ON  (unsigned int) 0x10
#define STATE_OFF (unsigned int) 0x20

//PB definitions
#define PB_POOL_RATE (unsigned int) 100                  //100 ms pool rate on push button
#define PB_SHORT_PRESS_TIME (float) 1.5                  //1.5 sec or shorter is classified as short press
#define PB_LONG_PRESS_TIME (float) 10                    //TIME_PB_SHORT_PRESS< pressTime< 10s long press
#define PB_SHORT_PRESS (unsigned int) 0x10
#define PB_LONG_PRESS  (unsigned int) 0x20
#define PB_NO_PRESS    (unsigned int) 0x30


//Pin definitions
#define RELAY       (unsigned int) 4
#define LED         (unsigned int) 12
#define BTN         (unsigned int) 13
#define SENSE_PROXY (unsigned int) 14




//Global variables
ESP8266WebServer server(80);
Ticker ledTicker;
unsigned int socketState = STATE_OFF;
String friendlyName;
//////////////////////////////////



#define readBTNpres (digitalRead(BTN) ? false : true)

void LedOn()  { digitalWrite(LED, LOW); }
void LedOff() { digitalWrite(LED, HIGH); }
void LedTick(){ digitalWrite(LED, !digitalRead(LED));}

void RelayOn() { digitalWrite(RELAY, HIGH); }
void RelayOff() { digitalWrite(RELAY, LOW); }


void initPerip (){
  //Init GPIOs direction
    pinMode(RELAY, OUTPUT);
    pinMode(LED, OUTPUT);
    pinMode(BTN, INPUT);
    pinMode(SENSE_PROXY, INPUT);
    
    //initial periph. states
    LedOn();
    RelayOff();
}

void wifiConfigModeCallback(WiFiManager *myWiFiManager) {
  Serial.println("Entered config mode");
  ledTicker.attach(1, LedTick);
}

void setState(unsigned int _state){
    if (_state == STATE_ON || _state == STATE_OFF){
      socketState = _state;
    }
}


unsigned int getState(void){
  return socketState;
}

void httpRequest(){
  LedOn();
  int n = server.args();
  if (n) {
      String _argName = server.argName(0);
      String _argVal =  server.arg(0);
      if (_argName == "json") {
          StaticJsonBuffer<200> jsonBuffer;
          JsonObject& root = jsonBuffer.parseObject(_argVal);
          String state = root["state"];
          if (state == "on"){
            setState(STATE_ON);
          }else if (state == "off"){
            setState(STATE_OFF);
          }else{
            //error, reload page
          }
      }else{
          //error, reload page
      }    
  }
  /*if (getState() == STATE_ON){
    server.send(200, "text/html", htmlMessage);
  }else{
    server.send(200, "text/plain", "State == OFF");
  }*/

  // Display the HTML web page
 String htmlMsg = "<!DOCTYPE html><html>\n";
        htmlMsg += "<head><meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">\n";
        htmlMsg += "<link rel=\"icon\" href=\"data:,\">\n";
        // CSS to style the on/off buttons 
        // Feel free to change the background-color and font-size attributes to fit your preferences
        htmlMsg += "<style>html { font-family: Helvetica; display: inline-block; margin: 0px auto; text-align: center;}\n";
        htmlMsg += ".button { background-color: green; border: none; color: white; padding: 16px 40px;\n";
        htmlMsg += "text-decoration: none; font-size: 60px; margin: 2px; cursor: pointer;}\n";
        htmlMsg += ".button2 {background-color: red;}</style></head>\n";
        
        // Web Page Heading
        htmlMsg += "<body><h1>Pametna Uticnica</h1>\n";
        htmlMsg += "<body><h1>Ime: " + friendlyName + "</h1>\n";

        if (getState() == STATE_ON && digitalRead(SENSE_PROXY)){
          // Display current state, and ON/OFF buttons for GPIO 5  
          htmlMsg += "<p> Utikac: prisutan </p>\n";  
          htmlMsg += "<p> Stanje uticnice: Ukljuceno </p>\n";  
          htmlMsg += "<p><a href=\"/?json={%22state%22:%22off%22}\"><button class=\"button button2\">OFF</button></a></p>\n";
        }else{
          //current state = off
          htmlMsg += digitalRead(SENSE_PROXY)==HIGH ?  "<p> Utikac: prisutan </p>\n" : "<p> Utikac : nije prisutan </p>\n" ;
          htmlMsg += "<p> Stanje uticnice: Iskljuceno</p>\n";  
          htmlMsg += "<p><a href=\"/?json={%22state%22:%22on%22}\"><button class=\"button\">ON</button></a></p>\n";
        }
        htmlMsg += "</body></html>\n";
          
        // The HTTP response ends with another blank line
        htmlMsg +="\n";

        server.send(200, "text/html", htmlMsg);
        
  LedOff();
}

class http_task : public Task {
   protected:
      void setup() {}
      void loop(){
        server.handleClient(); //Handling of incoming requests
        delay(100);
      }
  
}httpTask;

class socket_agent_task : public Task {
   protected:
      void setup() {}
      void loop(){
        socketAgent();
        delay(100);
      }
  
}socketAgentTask;

class pbDetectTask : public Task {
  protected:
  void setup() {
    cnt = 0;
    pbState = PB_NO_PRESS;
  }
  void loop() {
    if (readBTNpres){
      //btn is pressed
      cnt++;
    }else{
      if (cnt){
        if (cnt > PB_SHORT_PRESS_TIME * 1000/PB_POOL_RATE){
          if (cnt <= PB_LONG_PRESS_TIME * 1000/PB_POOL_RATE){
            //long press occured
            pbState = PB_LONG_PRESS;
          }else{
            //error timeout occured
            pbState = PB_NO_PRESS;
          }
        }else{
          //short press occured
          pbState = PB_SHORT_PRESS;
          
        }
      cnt = 0;
      }
    }
    delay(100);
  }
  private:
    uint8_t cnt;
    uint8_t pbState;
   public:
     unsigned int getPBstate(void){
        unsigned int temp = (unsigned int) pbState;
        pbState = PB_NO_PRESS;
        return temp;
   }
} pbTask;

void socketAgent (){
  
  if (digitalRead(SENSE_PROXY)){
  //if (1){
    //ostavi/promjeni stanje
    if (socketState == STATE_ON){
      if (pbTask.getPBstate() == PB_LONG_PRESS){
        socketState = STATE_OFF;
      }
    }else{
      if (pbTask.getPBstate() == PB_SHORT_PRESS){
        socketState = STATE_ON;
      }
    }
  }else{
    //reset to default state
    if (socketState != STATE_OFF){
      socketState = STATE_OFF;
    }
  }

  if (socketState == STATE_ON){
    RelayOn();
    LedOn();
  }else{
    RelayOff();
    LedOff();
  }
}

void setup() {
    //serial debug interface
    Serial.begin(115200);

    //init Peripherals
    initPerip();

    //read frindlyName from memory
    if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);
        DynamicJsonBuffer jsonBuffer;
        JsonObject& json = jsonBuffer.parseObject(buf.get());
        json.printTo(Serial);
        if (json.success()) {
          Serial.println("\nparsed json");
          String fn = json["friendlyName"];
          friendlyName = fn;
        } else {
          Serial.println("failed to load json config");
        }
      }
    }
  } else {
    Serial.println("failed to mount FS");
  }
  
    //Wifi setup
    WiFiManager wifiManager;
    wifiManager.setAPCallback(wifiConfigModeCallback);
    
    byte mac [6];
    WiFi.macAddress(mac);
    friendlyName = "S2U2_" + String(mac[4], HEX) + String(mac[5], HEX) + String(mac[6], HEX);
    WiFiManagerParameter hostName("server", "hostname", friendlyName.c_str(), 40);
    wifiManager.addParameter(&hostName);
    
    
    for (int i=0; i<6; i++) Serial.println(mac[i]);
    String ssid =  friendlyName;   
    char ssid_c [50] = {0};
    ssid.toCharArray(ssid_c, 50);
    if (!wifiManager.autoConnect(ssid_c, "12345678")){
      Serial.println("Connection timeout, RESET");
      ESP.reset();
    }
    //signal succesful wifi connection
    friendlyName = String(hostName.getValue());
    //save to flash
    Serial.println("saving config");
    DynamicJsonBuffer jsonBuffer;
    JsonObject& json = jsonBuffer.createObject();
    json["friendlyName"] = friendlyName;

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    json.printTo(Serial);
    json.printTo(configFile);
    configFile.close();
    //end save
    
    ledTicker.attach(0.2, LedTick);
    delay(1000);
    ledTicker.detach();
    LedOff();
    // print ip address for debuging
    Serial.println("ip addr");
    Serial.println(WiFi.localIP());
    //Wifi setup complete


    //mDNS setup
    MDNS.begin(friendlyName.c_str());   //pitaj za hostname u wifi manageru



    server.on("/", httpRequest);
    /*server.on("/", [](){
      server.send(200, "text/html", htmlMessage);
    });

    server.on("/javascript", []() { //Define the handling function for the javascript path
      server.send(200, "text/html", javascriptCode);
    });

     server.on("/cssButton", []() { //Define the handling function for the CSS path
       server.send(200, "text/html", cssButton);
     });

     server.on("/led", httpRequestLed);
     
     */
     
      server.begin(); //Start the server

      Serial.println("Server listening");

      Scheduler.start(&pbTask);
      Scheduler.start(&httpTask);
      Scheduler.start(&socketAgentTask);
      Scheduler.begin();//start scheduler
}

void loop() {
      //server.handleClient(); //Handling of incoming requests
      //socketAgent();
      
}
